"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const typescript_estree_1 = require("@typescript-eslint/typescript-estree");
const tsutils = __importStar(require("tsutils"));
const ts = __importStar(require("typescript"));
const util = __importStar(require("../util"));
exports.default = util.createRule({
    name: 'unbound-method',
    meta: {
        docs: {
            category: 'Best Practices',
            description: 'Enforces unbound methods are called with their expected scope.',
            tslintName: 'no-unbound-method',
            recommended: 'error',
        },
        messages: {
            unbound: 'Avoid referencing unbound methods which may cause unintentional scoping of `this`.',
        },
        schema: [
            {
                type: 'object',
                properties: {
                    ignoreStatic: {
                        type: 'boolean',
                    },
                },
                additionalProperties: false,
            },
        ],
        type: 'problem',
    },
    defaultOptions: [
        {
            ignoreStatic: false,
        },
    ],
    create(context, [{ ignoreStatic }]) {
        const parserServices = util.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        return {
            [typescript_estree_1.AST_NODE_TYPES.MemberExpression](node) {
                if (isSafeUse(node)) {
                    return;
                }
                const originalNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const symbol = checker.getSymbolAtLocation(originalNode);
                if (symbol && isDangerousMethod(symbol, ignoreStatic)) {
                    context.report({
                        messageId: 'unbound',
                        node,
                    });
                }
            },
        };
    },
});
function isDangerousMethod(symbol, ignoreStatic) {
    const { valueDeclaration } = symbol;
    switch (valueDeclaration.kind) {
        case ts.SyntaxKind.MethodDeclaration:
        case ts.SyntaxKind.MethodSignature:
            return !(ignoreStatic &&
                tsutils.hasModifier(valueDeclaration.modifiers, ts.SyntaxKind.StaticKeyword));
    }
    return false;
}
function isSafeUse(node) {
    const parent = node.parent;
    switch (parent.type) {
        case typescript_estree_1.AST_NODE_TYPES.IfStatement:
        case typescript_estree_1.AST_NODE_TYPES.ForStatement:
        case typescript_estree_1.AST_NODE_TYPES.MemberExpression:
        case typescript_estree_1.AST_NODE_TYPES.UpdateExpression:
        case typescript_estree_1.AST_NODE_TYPES.WhileStatement:
            return true;
        case typescript_estree_1.AST_NODE_TYPES.CallExpression:
            return parent.callee === node;
        case typescript_estree_1.AST_NODE_TYPES.ConditionalExpression:
            return parent.test === node;
        case typescript_estree_1.AST_NODE_TYPES.LogicalExpression:
            return parent.operator !== '||';
        case typescript_estree_1.AST_NODE_TYPES.TaggedTemplateExpression:
            return parent.tag === node;
        case typescript_estree_1.AST_NODE_TYPES.TSNonNullExpression:
        case typescript_estree_1.AST_NODE_TYPES.TSAsExpression:
        case typescript_estree_1.AST_NODE_TYPES.TSTypeAssertion:
            return isSafeUse(parent);
    }
    return false;
}
//# sourceMappingURL=unbound-method.js.map